from __future__ import annotations

from ._sqlmodel_load_from_db import SQLModelLoadFromDb
from ._sqlmodel_db_connection import SQLModelConnectionConfig

__all__ = [
    "SQLModelLoadFromDb",
    "SQLModelConnectionConfig",
]



import itertools
import dask.dataframe as dd
import pandas as pd

from django.core.cache import cache
from django.db import models
from django.db.models import Field
from django.utils.encoding import force_str as force_text
import django
from django.core.exceptions import FieldDoesNotExist

class ReadFrameDask:
    FieldDoesNotExist = (
        django.core.exceptions.FieldDoesNotExist
        if django.VERSION < (1, 8)
        else django.core.exceptions.FieldDoesNotExist
    )

    def __init__(
            self,
            qs,
            **kwargs,
    ):
        self.qs = qs
        self.coerce_float = kwargs.setdefault("coerce_float", False)
        self.chunk_size = kwargs.setdefault("chunk_size", 1000)
        self.verbose = kwargs.setdefault("verbose", True)

    @staticmethod
    def get_model_name(model):
        return model._meta.model_name

    @staticmethod
    def get_related_model(field):
        model = None
        if hasattr(field, "related_model") and field.related_model:
            model = field.related_model
        elif hasattr(field, "rel") and field.rel:
            model = field.rel.to
        return model

    @classmethod
    def get_base_cache_key(cls, model):
        return (
            f"dask_{model._meta.app_label}_{cls.get_model_name(model)}_%s_rendering"
        )

    @classmethod
    def replace_pk(cls, model):
        base_cache_key = cls.get_base_cache_key(model)

        def get_cache_key_from_pk(pk):
            return None if pk is None else base_cache_key % str(pk)

        def inner(pk_series):
            pk_series = pk_series.astype(object).where(pk_series.notnull(), None)
            cache_keys = pk_series.apply(get_cache_key_from_pk, convert_dtype=False)
            unique_cache_keys = list(filter(None, cache_keys.unique()))
            if not unique_cache_keys:
                return pk_series

            out_dict = cache.get_many(unique_cache_keys)
            if len(out_dict) < len(unique_cache_keys):
                out_dict = dict(
                    [
                        (base_cache_key % obj.pk, force_text(obj))
                        for obj in model.objects.filter(
                        pk__in=list(filter(None, pk_series.unique()))
                    )
                    ]
                )
                cache.set_many(out_dict)
            return list(map(out_dict.get, cache_keys))

        return inner

    @staticmethod
    def replace_from_choices(choices):
        def inner(values):
            return [choices.get(v, v) for v in values]

        return inner

    @classmethod
    def build_update_functions(cls, fieldnames, fields):
        for fieldname, field in zip(fieldnames, fields):
            if not isinstance(field, Field):
                yield fieldname, None
            else:
                if field.choices:
                    choices = dict([(k, force_text(v)) for k, v in field.flatchoices])
                    yield fieldname, cls.replace_from_choices(choices)
                elif field.get_internal_type() == "ForeignKey":
                    yield fieldname, cls.replace_pk(cls.get_related_model(field))

    @classmethod
    def update_with_verbose(cls, df, fieldnames, fields):
        for fieldname, function in cls.build_update_functions(fieldnames, fields):
            if function is not None:
                df[fieldname] = df[fieldname].map_partitions(lambda x: function(x))

    @staticmethod
    def infer_dtypes_from_django(qs):
        """Infers Dask data types based on Django queryset model fields, with support for nullable integers."""
        django_to_dask_dtype = {
            'AutoField': 'Int64',  # Use nullable integer
            'BigAutoField': 'Int64',
            'BigIntegerField': 'Int64',
            'BooleanField': 'bool',
            'CharField': 'object',
            'DateField': 'datetime64[ns]',
            'DateTimeField': 'datetime64[ns]',
            'DecimalField': 'float64',
            'FloatField': 'float64',
            'IntegerField': 'Int64',  # Use nullable integer
            'PositiveIntegerField': 'Int64',
            'SmallIntegerField': 'Int64',
            'TextField': 'object',
            'TimeField': 'object',
            'UUIDField': 'object',
            'ForeignKey': 'Int64',  # Use nullable integer for FK fields
        }

        dtypes = {}
        # Handle model fields
        for field in qs.model._meta.get_fields():
            # Skip reverse relationships and non-concrete fields
            if not getattr(field, 'concrete', False):
                continue

            # Check for AutoField or BigAutoField explicitly
            if isinstance(field, (models.AutoField, models.BigAutoField)):
                dtypes[field.name] = 'Int64'  # Nullable integer for autoincremented fields
            else:
                # Use field type to infer dtype
                field_type = field.get_internal_type()
                dtypes[field.name] = django_to_dask_dtype.get(field_type, 'object')

        # Handle annotated fields
        for annotation_name, annotation in qs.query.annotation_select.items():
            if hasattr(annotation, 'output_field'):
                field_type = annotation.output_field.get_internal_type()
                dtype = django_to_dask_dtype.get(field_type, 'object')
            else:
                dtype = 'object'  # Default to object for untyped annotations
            dtypes[annotation_name] = dtype

        return dtypes

    def read_frame(self, fillna_value=None):
        qs = self.qs
        fieldnames = tuple(qs.model._meta.get_fields())
        dtypes = self.infer_dtypes_from_django(qs)
        chunk_size = self.chunk_size
        verbose = self.verbose

        # Use values to directly fetch required fields
        qs = qs.values(*fieldnames)

        # Create partitions for Dask
        partitions = []
        iterator = qs.iterator(chunk_size=chunk_size)
        for chunk in itertools.islice(iterator, chunk_size):
            df = pd.DataFrame.from_records(chunk, columns=fieldnames)

            # Handle NaN values
            if fillna_value:
                df = df.fillna(fillna_value)

            # Optimize timezone conversions
            for col in df.columns:
                if isinstance(df[col].dtype, pd.DatetimeTZDtype):
                    df[col] = df[col].dt.tz_localize(None)

            # Optimize dtype conversion
            df = df.convert_dtypes()

            # Convert to Dask DataFrame
            partitions.append(dd.from_pandas(df, npartitions=1))

        # Combine all partitions
        dask_df = dd.concat(partitions, axis=0, ignore_index=True)

        # Apply verbose updates
        if verbose:
            self.update_with_verbose(dask_df, fieldnames, qs.model._meta.fields)

        return dask_df


import asyncio
import datetime
from typing import Any, Dict, TypeVar
from typing import Union, Optional

import dask.dataframe as dd
import pandas as pd
from pydantic import BaseModel

from sibi_dst.utils import ParquetSaver, ClickHouseWriter
from sibi_dst.df_helper.core import QueryConfig, ParamsConfig
from sibi_dst.utils import Logger
from .plugins.django import *
from .plugins.http import HttpConfig
from .plugins.parquet import ParquetConfig, ParquetFilterHandler
from .plugins.sql_alchemy import *

# Define a generic type variable for BaseModel subclasses
T = TypeVar("T", bound=BaseModel)

class DfHelper:
    df: Union[dd.DataFrame, pd.DataFrame] = None
    plugin_django_connection: Optional[DjangoConnectionConfig] = None
    plugin_query: Optional[QueryConfig] = None
    plugin_params: Optional[ParamsConfig] = None
    plugin_parquet: Optional[ParquetConfig] = None
    plugin_http: Optional[HttpConfig] = None
    plugin_sqlalchemy: Optional[SqlAlchemyConnectionConfig] = None
    parquet_filename: str = None
    logger: Logger
    default_config: Dict = None

    def __init__(self, source='django_db', **kwargs):
        # Ensure default_config is not shared across instances
        self.default_config = self.default_config or {}
        kwargs = {**self.default_config.copy(), **kwargs}
        self.source = source
        self.logger = Logger.default_logger(logger_name=self.__class__.__name__)
        self.debug = kwargs.setdefault("debug", False)
        self.verbose_debug = kwargs.setdefault("verbose_debug", False)
        self.parquet_storage_path = kwargs.setdefault("parquet_storage_path", None)
        self.dt_field=kwargs.setdefault("dt_field", None)
        self.as_pandas = kwargs.setdefault("as_pandas", False)
        kwargs.setdefault("live", True)
        kwargs.setdefault("logger", self.logger)
        self.post_init(**kwargs)


    def post_init(self, **kwargs):
        self.logger.info(f"Source used: {self.source}")
        self.plugin_query = self.__get_config(QueryConfig, kwargs)
        self.plugin_params = self.__get_config(ParamsConfig, kwargs)
        if self.source == 'django_db':
            self.plugin_django_connection = self.__get_config(DjangoConnectionConfig, kwargs)
        elif self.source == 'parquet':
            self.parquet_filename = kwargs.setdefault("parquet_filename", None)
            self.plugin_parquet = ParquetConfig(**kwargs)
        elif self.source == 'http':
            self.plugin_http = HttpConfig(**kwargs)
        elif self.source == 'sqlalchemy':
            self.plugin_sqlalchemy = self.__get_config(SqlAlchemyConnectionConfig,kwargs)

    @staticmethod
    def __get_config(model: [T], kwargs: Dict[str, Any]) -> Union[T]:
        """
        Initializes a Pydantic model with the keys it recognizes from the kwargs,
        and removes those keys from the kwargs dictionary.
        :param model: The Pydantic model class to initialize.
        :param kwargs: The dictionary of keyword arguments.
        :return: The initialized Pydantic model instance.
        """
        # Extract keys that the model can accept
        recognized_keys = set(model.__annotations__.keys())
        model_kwargs = {k: kwargs.pop(k) for k in list(kwargs.keys()) if k in recognized_keys}
        return model(**model_kwargs)

    def load(self, **options):
        # this will be the universal method to load data from a df irrespective of the source
        df = self._load(**options)
        if self.as_pandas:
            return df.compute()
        return df

    def _load(self, **options):

        if self.source == 'django_db':
            self.plugin_params.parse_params(options)
            return self._load_from_db(**options)
        elif self.source == 'sqlalchemy':
            self.plugin_params.parse_params(options)
            return self._load_from_sqlalchemy(**options)
        elif self.source == 'parquet':
            return self._load_from_parquet(**options)
        elif self.source == 'http':
            if asyncio.get_event_loop().is_running():
                self.logger.info("Running as a task from an event loop")
                return asyncio.create_task(self._load_from_http(**options))
            else:
                self.logger.info("Regular asyncio run...")
                return asyncio.run(self._load_from_http(**options))


    def _load_from_sqlalchemy(self, **options):
        try:
            options.setdefault("debug", self.debug)
            options.setdefault("verbose_debug", self.verbose_debug)
            db_loader = SqlAlchemyLoadFromDb(
                self.plugin_sqlalchemy,
                self.plugin_query,
                self.plugin_params,
                self.logger,
                **options
            )
            self.df = db_loader.build_and_load()
            self._process_loaded_data()
            self._post_process_df()
            self.logger.info("Data successfully loaded from sqlalchemy database.")
        except Exception as e:
            self.logger.error(f"Failed to load data from sqlqlchemy database: {e}")
            self.df = dd.from_pandas(pd.DataFrame(), npartitions=1)

        return self.df

    def _load_from_db(self, **options) -> Union[pd.DataFrame, dd.DataFrame]:
        try:
            options.setdefault("debug", self.debug)
            options.setdefault("verbose_debug", self.verbose_debug)
            db_loader = DjangoLoadFromDb(
                self.plugin_django_connection,
                self.plugin_query,
                self.plugin_params,
                self.logger,
                **options
            )
            self.df = db_loader.build_and_load()
            self._process_loaded_data()
            self._post_process_df()
            self.logger.info("Data successfully loaded from django database.")
        except Exception as e:
            self.logger.error(f"Failed to load data from django database: {e}")
            self.df=dd.from_pandas(pd.DataFrame(), npartitions=1)

        return self.df

    async def _load_from_http(self, **options) -> Union[pd.DataFrame, dd.DataFrame]:
        """Delegate asynchronous HTTP data loading to HttpDataSource plugin."""
        if not self.plugin_http:
            self.logger.error("HTTP plugin not configured properly.")
            return dd.from_pandas(pd.DataFrame(), npartitions=1)
        try:
            self.df = await self.plugin_http.fetch_data(**options)
        except Exception as e:
            self.logger.error(f"Failed to load data from http plugin: {e}")
            self.df=dd.from_pandas(pd.DataFrame(), npartitions=1)
        return self.df


    def _post_process_df(self):
        """
        Efficiently process the DataFrame by filtering, renaming, and setting indices.
        Optimized for large datasets with Dask compatibility.
        """
        df_params = self.plugin_params.df_params
        fieldnames = df_params.get("fieldnames", None)
        index_col = df_params.get("index_col", None)
        datetime_index = df_params.get("datetime_index", False)
        column_names = df_params.get("column_names", None)

        # Filter columns
        if fieldnames:
            existing_columns = set(self.df.columns)
            valid_fieldnames = list(filter(existing_columns.__contains__, fieldnames))
            self.df = self.df[valid_fieldnames]

        # Rename columns
        if column_names is not None:
            if len(fieldnames) != len(column_names):
                raise ValueError(
                    f"Length mismatch: fieldnames ({len(fieldnames)}) and column_names ({len(column_names)}) must match."
                )
            rename_mapping = dict(zip(fieldnames, column_names))
            self.df = self.df.map_partitions(lambda df: df.rename(columns=rename_mapping))

        # Set index column
        if index_col is not None:
            if index_col in self.df.columns:
                self.df = self.df.set_index(index_col)
            else:
                raise ValueError(f"Index column '{index_col}' not found in DataFrame.")

        # Handle datetime index
        if datetime_index and self.df.index.dtype != 'datetime64[ns]':
            self.df = self.df.map_partitions(lambda df: df.set_index(pd.to_datetime(df.index, errors='coerce')))

        self.logger.info("Post-processing of DataFrame completed.")

    def _process_loaded_data(self):
        if len(self.df.index) > 0:
            field_map = self.plugin_params.field_map or []
            if field_map:
                rename_mapping = {k: v for k, v in field_map.items() if k in self.df.columns}

                def rename_columns(df, mapping):
                    return df.rename(columns=mapping)

                if rename_mapping:
                    # Apply renaming
                    self.df = self.df.map_partitions(rename_columns, mapping=rename_mapping)

            self.logger.info("Processing of loaded data completed.")

    def save_to_parquet(self, parquet_filename: Optional[str] = None):
        ps = ParquetSaver(self.df, self.parquet_storage_path, self.logger)
        ps.save_to_parquet(parquet_filename)
        self.logger.info(f"Parquet saved to {parquet_filename} in parquet storage: {self.parquet_storage_path}.")

    def save_to_clickhouse(self, database, table, order_by=None, **credentials):
        click_config ={
            'database': database,
            'table': table,
            'order_by': order_by or 'id',
        }
        credentials = {**credentials, **click_config}
        cs=ClickHouseWriter(logger=self.logger, **credentials)
        cs.save_to_clickhouse(self.df)
        self.logger.info("Save to ClickHouse completed.")

    def _load_from_parquet(self, **options) -> Union[pd.DataFrame, dd.DataFrame]:
        self.df = self.plugin_parquet.load_files()
        if options:
            self.df = ParquetFilterHandler(logger=self.logger).apply_filters_dask(self.df, options)
        return self.df

    def load_period(self, **kwargs):
        return self.__load_period(**kwargs)

    def __load_period(self, **kwargs):
        dt_field = kwargs.pop("dt_field", self.dt_field)
        if dt_field is None:
            raise ValueError("dt_field must be provided")

        start = kwargs.pop("start", None)
        end = kwargs.pop("end", None)

        # Ensure start and end are provided
        if start is None or end is None:
            raise ValueError("Both 'start' and 'end' must be provided.")

        # Parse string dates
        if isinstance(start, str):
            start = self.parse_date(start)
        if isinstance(end, str):
            end = self.parse_date(end)

        # Validate that start <= end
        if start > end:
            raise ValueError("The 'start' date cannot be later than the 'end' date.")

        # Reverse map to original field name
        field_map = getattr(self.plugin_params, 'field_map', {}) or {}
        reverse_map = {v: k for k, v in field_map.items()}
        mapped_field = reverse_map.get(dt_field, dt_field)

        # Common logic for Django and SQLAlchemy
        if self.source == 'django_db':
            model_fields = {field.name: field for field in self.plugin_django_connection.model._meta.get_fields()}
            if mapped_field not in model_fields:
                raise ValueError(f"Field '{dt_field}' does not exist in the Django model.")
            field_type = type(model_fields[mapped_field]).__name__
            is_date_field = field_type == 'DateField'
            is_datetime_field = field_type == 'DateTimeField'
        elif self.source == 'sqlalchemy':
            model = self.plugin_sqlalchemy.model
            fields = [column.name for column in model.__table__.columns]
            if mapped_field not in fields:
                raise ValueError(f"Field '{dt_field}' does not exist in the SQLAlchemy model.")
            column = getattr(model, mapped_field)
            field_type = str(column.type).upper()
            is_date_field = field_type == 'DATE'
            is_datetime_field = field_type == 'DATETIME'
        else:
            raise ValueError(f"Unsupported source '{self.source}'")
            # Build query filters
        if start == end:
            if is_date_field:
                kwargs[mapped_field] = start
            elif is_datetime_field:
                kwargs[f"{mapped_field}__date"] = start
        else:
            if is_date_field:
                kwargs[f"{mapped_field}__gte"] = start
                kwargs[f"{mapped_field}__lte"] = end
            elif is_datetime_field:
                kwargs[f"{mapped_field}__date__gte"] = start
                kwargs[f"{mapped_field}__date__lte"] = end
        return self.load(**kwargs)


    @staticmethod
    def parse_date(date_str: str) -> Union[datetime.datetime, datetime.date]:
        try:
            return datetime.datetime.strptime(date_str, '%Y-%m-%d %H:%M:%S')
        except ValueError:
            return datetime.datetime.strptime(date_str, '%Y-%m-%d').date()



from typing import Dict

import dask.dataframe as dd
import pandas as pd
from sqlalchemy.inspection import inspect
from sqlalchemy.orm import sessionmaker
from sqlalchemy import select
#from sqlmodel import Session, select

from sibi_dst.df_helper.core import ParamsConfig, QueryConfig, sqlalchemy_field_conversion_map_dask, \
    normalize_sqlalchemy_type
from sibi_dst.utils import Logger
from ._io_sqlalchemy_dask import SQLAlchemyDask
from ._sqlachemy_filter_handler import SqlAlchemyFilterHandler
from ._sqlalchemy_db_connection import SqlAlchemyConnectionConfig


class SqlAlchemyLoadFromDb:
    df: dd.DataFrame

    def __init__(
            self,
            plugin_sqlalchemy: SqlAlchemyConnectionConfig,  # Expected to be an instance of SqlAlchemyConnection
            plugin_query: QueryConfig = None,
            plugin_params: ParamsConfig = None,
            logger: Logger = None,
            **kwargs,
    ):
        """
        Initialize the loader with database connection, query, and parameters.
        """
        self.db_connection = plugin_sqlalchemy
        self.table_name = self.db_connection.table
        self.model = self.db_connection.model
        self.engine = self.db_connection.engine
        self.logger = logger or Logger.default_logger(logger_name=self.__class__.__name__)
        self.query_config = plugin_query
        self.params_config = plugin_params
        self.debug = kwargs.pop("debug", False)
        self.verbose_debug = kwargs.pop("verbose_debug", False)

    def build_and_load(self) -> dd.DataFrame:
        """
        Load data into a Dask DataFrame based on the query and parameters.
        """
        self.df = self._build_and_load()
        return self.df

    def _build_and_load(self) -> dd.DataFrame:
        try:
            reader = SQLAlchemyDask(model=self.model, filters=self.params_config.filters,engine_url=self.engine.url, logger=self.logger, chunk_size=1000, verbose=self.debug)
            df = reader.read_frame()
            if df is None or len(df.index) == 0:
                self.logger.warning("Query returned no results.")
                return dd.from_pandas(pd.DataFrame(), npartitions=1)
            return df
        except Exception as e:
            self.logger.error(f"Failed to load data into Dask DataFrame.{e}")
            return dd.from_pandas(pd.DataFrame(), npartitions=1)
